/************************************************************************/
/* pnetbuffer.h								*/
/************************************************************************/
/*
 *  (c)2017,2020-2021 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/




#ifndef _PNETBUFFER_H_
#define _PNETBUFFER_H_

#include <stdint.h>


// If defined, inline most funtions which can
// provide upto 6x speed improvements
#define PNETBUFFER_INLINE_MOST



typedef enum {
	PNETBUFFER_RC_SUCCESS	= 0,
	PNETBUFFER_RC_NOSPACE,
	PNETBUFFER_RC_NOTENOUGH,
	PNETBUFFER_RC_MISSINGOPT,
} pnetbuffer_rc_t;


#define PNETBUFFER_OPT_STRINGS_COUNT_FIRST      0x01
#define PNETBUFFER_OPT_STRINGS_LAST_EMPTY	0x02

// Opaque prototype
typedef struct pnetbuffer_s	pnetbuffer_t;




#ifndef PNETBUFFER_INLINE_MOST

pnetbuffer_t *pnetbuffer_init(char *buffer, uint16_t len);
void pnetbuffer_reset(pnetbuffer_t *nb);
void pnetbuffer_reset_read(pnetbuffer_t *nb);

// add functions take LSB values and writes to the buffer in MSB
uint8_t pnetbuffer_add_uint8(pnetbuffer_t *nb, uint8_t val);
uint8_t pnetbuffer_add_uint16(pnetbuffer_t *nb, uint16_t val);
uint8_t pnetbuffer_add_uint32(pnetbuffer_t *nb, uint32_t val);
uint8_t pnetbuffer_add_uint64(pnetbuffer_t *nb, uint64_t val);
uint8_t pnetbuffer_add_double(pnetbuffer_t *nb, double val);
uint8_t pnetbuffer_add_float(pnetbuffer_t *nb, float val);
uint8_t pnetbuffer_add_data(pnetbuffer_t *nb, char *data, uint16_t datalen);
uint8_t pnetbuffer_add_string(pnetbuffer_t *nb, char *str);
uint8_t pnetbuffer_add_strings(pnetbuffer_t *nb, char **strings, uint8_t count, uint8_t options);


// read functions take MSB values from the buffer and return LSB values
uint8_t pnetbuffer_read_uint8(pnetbuffer_t *nb, uint8_t *val);
uint8_t pnetbuffer_read_uint16(pnetbuffer_t *nb, uint16_t *val);
uint8_t pnetbuffer_read_uint32(pnetbuffer_t *nb, uint32_t *val);
uint8_t pnetbuffer_read_uint64(pnetbuffer_t *nb, uint64_t *val);
uint8_t pnetbuffer_read_double(pnetbuffer_t *nb, double *val);
uint8_t pnetbuffer_read_float(pnetbuffer_t *nb, float *val);
uint8_t pnetbuffer_read_string(pnetbuffer_t *nb, char *str, uint16_t strlen);
uint8_t pnetbuffer_read_data(pnetbuffer_t *nb, char *data, uint16_t datalen);
uint8_t pnetbuffer_read_strings(pnetbuffer_t *nb,
				char **array,		// array allocated by caller
				uint8_t maxlen,		// max length of array elements
				uint8_t maxcount,	// maximum array length
				uint8_t *count,		// returned count of array elements filled
				uint8_t options);	// processing options

void pnetbuffer_get_readable(pnetbuffer_t *nb, char **ptr, uint16_t *len);
uint8_t pnetbuffer_read_advance(pnetbuffer_t *nb, uint16_t datalen);
uint16_t pnetbuffer_get_writable_len(pnetbuffer_t *nb);

void pnetbuffer_get_writable(pnetbuffer_t *nb, char **ptr, uint16_t *len);
uint8_t pnetbuffer_set_written(pnetbuffer_t *nb, uint16_t len);

// write functions take LSB values and writes to the buffer in MSB
// at the given offset into the buffer
uint8_t pnetbuffer_write_uint8(pnetbuffer_t *nb, uint16_t offset, uint8_t val);
uint8_t pnetbuffer_write_uint16(pnetbuffer_t *nb, uint16_t offset, uint16_t val);
uint8_t pnetbuffer_write_uint32(pnetbuffer_t *nb, uint16_t offset, uint32_t val);
uint8_t pnetbuffer_write_float(pnetbuffer_t *nb, uint16_t offset, float val);
uint8_t pnetbuffer_write_data(pnetbuffer_t *nb, uint16_t offset, char *data, uint16_t datalen);

// peek functions take MSB values from the buffer and return LSB values,
// WITHOUT advancing the buffer (this is only useful to peek the first value
// in the buffer)
uint8_t pnetbuffer_peek_uint8(pnetbuffer_t *nb, uint8_t *val)
uint8_t pnetbuffer_peek_uint16(pnetbuffer_t *nb, uint16_t *val)
uint8_t pnetbuffer_peek_uint32(pnetbuffer_t *nb, uint32_t *val)
uint64_t pnetbuffer_peek_uint64(pnetbuffer_t *nb, uint64_t *val)

#else

/*************************************************************************
* Use lining if PNETBUFFER_INLINE_MOST is defined
*************************************************************************/
#include <string.h>

#ifndef NULL
#define NULL 0L
#endif


struct pnetbuffer_s {
	uint16_t	size;   // buf is size bytes long
	uint16_t	wpos;
	uint16_t	rpos;
	uint8_t		buf[]; // buf member will be size bytes long
};


typedef union floatint_u floatint_t;
union floatint_u {
	float f;
	uint32_t u32;
};


typedef union doubleint_u doubleint_t;
union doubleint_u {
	double d;
	uint64_t u64;
};


static inline void pnetbuffer_reset(pnetbuffer_t *nb)
{
	nb->wpos = 0;
	nb->rpos = 0;
	return;
}


static inline pnetbuffer_t *pnetbuffer_init(char *buffer, uint16_t len)
{
	pnetbuffer_t	*p = (pnetbuffer_t*)buffer;
	if (len < sizeof(pnetbuffer_t))
		return NULL;
	p->size = len - sizeof(pnetbuffer_t);
	pnetbuffer_reset(p);
	return p;
}


static inline void pnetbuffer_reset_read(pnetbuffer_t *nb)
{
	nb->rpos = 0;
	return;
}


// add functions take LSB values and writes to the buffer in MSB
static inline uint8_t pnetbuffer_add_uint8(pnetbuffer_t *nb, uint8_t val)
{
	if (nb->wpos == nb->size)
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = val;
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_uint16(pnetbuffer_t *nb, uint16_t val)
{
	if (nb->wpos >= (nb->size-1))
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = ((val>>8) & 0xff);
	nb->buf[nb->wpos++] = (val & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_uint32(pnetbuffer_t *nb, uint32_t val)
{
	if (nb->wpos >= (nb->size-3))
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = ((val>>24) & 0xff);
	nb->buf[nb->wpos++] = ((val>>16) & 0xff);
	nb->buf[nb->wpos++] = ((val>>8) & 0xff);
	nb->buf[nb->wpos++] = (val & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_uint64(pnetbuffer_t *nb, uint64_t val)
{
	if (nb->wpos >= (nb->size-7))
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = ((val>>56) & 0xff);
	nb->buf[nb->wpos++] = ((val>>48) & 0xff);
	nb->buf[nb->wpos++] = ((val>>40) & 0xff);
	nb->buf[nb->wpos++] = ((val>>32) & 0xff);
	nb->buf[nb->wpos++] = ((val>>24) & 0xff);
	nb->buf[nb->wpos++] = ((val>>16) & 0xff);
	nb->buf[nb->wpos++] = ((val>>8) & 0xff);
	nb->buf[nb->wpos++] = (val & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_double(pnetbuffer_t *nb, double val)
{
	doubleint_t     u;
	u.d = val;
	if (nb->wpos >= (nb->size-7))
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = ((u.u64>>56) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>48) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>40) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>32) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>24) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>16) & 0xff);
	nb->buf[nb->wpos++] = ((u.u64>>8) & 0xff);
	nb->buf[nb->wpos++] = (u.u64 & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_float(pnetbuffer_t *nb, float val)
{
	floatint_t u;
	u.f = val;
	if (nb->wpos >= (nb->size-3))
		return PNETBUFFER_RC_NOSPACE;
	nb->buf[nb->wpos++] = ((u.u32>>24) & 0xff);
	nb->buf[nb->wpos++] = ((u.u32>>16) & 0xff);
	nb->buf[nb->wpos++] = ((u.u32>>8) & 0xff);
	nb->buf[nb->wpos++] = (u.u32 & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_data(pnetbuffer_t *nb, char *data, uint16_t datalen)
{
	uint16_t	i;
	if (nb->wpos >= (nb->size-(datalen-1)))
		return PNETBUFFER_RC_NOSPACE;
	for (i=0 ; i<datalen ; i++)
		nb->buf[nb->wpos++] = data[i];
	return PNETBUFFER_RC_SUCCESS;
}



// read functions take MSB values from the buffer and return LSB values
static inline uint8_t pnetbuffer_read_uint8(pnetbuffer_t *nb, uint8_t *val)
{
	if (nb->rpos >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	*val = nb->buf[nb->rpos++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_uint16(pnetbuffer_t *nb, uint16_t *val)
{
	if ((nb->rpos+1) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	*val = (nb->buf[nb->rpos++]&0xff)<<8;
	*val += nb->buf[nb->rpos++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_uint32(pnetbuffer_t *nb, uint32_t *val)
{
	if ((nb->rpos+3) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	*val = (nb->buf[nb->rpos++])<<24;
	*val += (nb->buf[nb->rpos++])<<16;
	*val += (nb->buf[nb->rpos++])<<8;
	*val += nb->buf[nb->rpos++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_uint64(pnetbuffer_t *nb, uint64_t *val)
{
	if ((nb->rpos+7) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	*val = ((uint64_t)nb->buf[nb->rpos++])<<56;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<48;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<40;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<32;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<24;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<16;
	*val += ((uint64_t)nb->buf[nb->rpos++])<<8;
	*val += (uint64_t)nb->buf[nb->rpos++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_double(pnetbuffer_t *nb, double *val)
{
	doubleint_t u;
	if ((nb->rpos+7) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	u.u64 = ((uint64_t)nb->buf[nb->rpos++])<<56;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<48;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<40;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<32;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<24;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<16;
	u.u64 += ((uint64_t)nb->buf[nb->rpos++])<<8;
	u.u64 += (uint64_t)nb->buf[nb->rpos++];
	*val = u.d;
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_float(pnetbuffer_t *nb, float *val)
{
	floatint_t u;
	if ((nb->rpos+3) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	u.u32 = (nb->buf[nb->rpos++])<<24;
	u.u32 += (nb->buf[nb->rpos++])<<16;
	u.u32 += (nb->buf[nb->rpos++])<<8;
	u.u32 += nb->buf[nb->rpos++];
	*val = u.f;
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_read_data(pnetbuffer_t *nb, char *data, uint16_t datalen)
{
	uint16_t	i;
	if ((nb->rpos+(datalen-1)) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	for (i=0 ; i<datalen ; i++)
		data[i] = nb->buf[nb->rpos++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline void pnetbuffer_get_readable(pnetbuffer_t *nb, char **ptr, uint16_t *len)
{
	if (ptr == 0L || len == 0L)
		return;
	*ptr = (char *)&nb->buf[nb->rpos];
	*len = nb->wpos - nb->rpos;
	return;
}


static inline uint8_t pnetbuffer_read_advance(pnetbuffer_t *nb, uint16_t datalen)
{
	if ((nb->rpos+(datalen-1)) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	nb->rpos += datalen;
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint16_t pnetbuffer_get_writable_len(pnetbuffer_t *nb)
{
	if (nb == 0L)
		return 0;
	return (nb->size - nb->wpos);
}


static inline void pnetbuffer_get_writable(pnetbuffer_t *nb, char **ptr, uint16_t *len)
{
	if (ptr == 0L || len == 0L)
		return;
	*ptr = (char *)&nb->buf[nb->wpos];
	*len = nb->size - nb->wpos;
	return;
}


static inline uint8_t pnetbuffer_set_written(pnetbuffer_t *nb, uint16_t len)
{
	if (nb->wpos >= (nb->size-(len-1)))
		return PNETBUFFER_RC_NOSPACE;
	nb->wpos += len;
	return PNETBUFFER_RC_SUCCESS;
}


// write functions take LSB values and writes to the buffer in MSB
// at the given offset into the buffer
static inline uint8_t pnetbuffer_write_uint8(pnetbuffer_t *nb, uint16_t offset, uint8_t val)
{
	if (offset > (nb->size-1))
		return PNETBUFFER_RC_NOTENOUGH;
	nb->buf[offset] = val;
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_write_uint16(pnetbuffer_t *nb, uint16_t offset, uint16_t val)
{
	if (offset > (nb->size-2))
		return PNETBUFFER_RC_NOTENOUGH;
	nb->buf[offset] = ((val>>8) & 0xff);
	nb->buf[offset+1] = (val & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_write_uint32(pnetbuffer_t *nb, uint16_t offset, uint32_t val)
{
	if (offset > (nb->size-4))
		return PNETBUFFER_RC_NOTENOUGH;
	nb->buf[offset] = ((val>>24) & 0xff);
	nb->buf[offset+1] = ((val>>16) & 0xff);
	nb->buf[offset+2] = ((val>>8) & 0xff);
	nb->buf[offset+3] = (val & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_write_float(pnetbuffer_t *nb, uint16_t offset, float val)
{
	floatint_t u;
	u.f = val;
	if (offset > (nb->size-4))
		return PNETBUFFER_RC_NOTENOUGH;
	nb->buf[offset] = ((u.u32>>24) & 0xff);
	nb->buf[offset+1] = ((u.u32>>16) & 0xff);
	nb->buf[offset+2] = ((u.u32>>8) & 0xff);
	nb->buf[offset+3] = (u.u32 & 0xff);
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_write_data(pnetbuffer_t *nb, uint16_t offset, char *data, uint16_t datalen)
{
	uint16_t	i;
	if (offset > (nb->size-datalen))
		return PNETBUFFER_RC_NOTENOUGH;
	for (i=0 ; i<datalen ; i++) {
		nb->buf[offset+i] = data[i];
	}
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_add_strings(pnetbuffer_t *nb, char **strings, uint8_t count, uint8_t options)
{
	uint8_t		rc;
	uint8_t		i;
	uint16_t	len;

	if (options & PNETBUFFER_OPT_STRINGS_COUNT_FIRST) {
		// count preceeds the strings (pre-knowledge)
		rc = pnetbuffer_add_uint8(nb, count);
		if (rc != 0) return rc;
	}
	for (i=0 ; i<count ; i++) {
		len = strlen(strings[i]);
		rc = pnetbuffer_add_uint8(nb, len);
		if (rc != 0) return rc;
		rc = pnetbuffer_add_data(nb, strings[i], len);
		if (rc != 0) return rc;
	}
	if (options & PNETBUFFER_OPT_STRINGS_LAST_EMPTY) {
		// append a zero length string (marks end of list)
		rc = pnetbuffer_add_uint8(nb, 0);
		if (rc != 0) return rc;
	}

	return 0;
}



static inline uint8_t pnetbuffer_read_strings(pnetbuffer_t *nb,
				char **array,		// array allocated by caller
				uint8_t maxlen,		// max length of array elements
				uint8_t maxcount,	// maximum array length
				uint8_t *count,		// returned count of array elements filled
				uint8_t options)	// processing options
{
	uint8_t		rc;
	uint8_t		i;
	uint8_t		nbcount;
	uint8_t		len;

	*count = 0;
	if (options & PNETBUFFER_OPT_STRINGS_COUNT_FIRST) {
		// count preceeds the strings (pre-knowledge)
		rc = pnetbuffer_read_uint8(nb, &nbcount);
		if (rc != 0) return rc;
		if (nbcount > maxcount) return PNETBUFFER_RC_NOSPACE;
		for (i=0; i<nbcount ; i++) {
			rc = pnetbuffer_read_uint8(nb, &len);
			if (rc != 0) return rc;
			if (len >= maxlen) return PNETBUFFER_RC_NOSPACE;
			rc = pnetbuffer_read_data(nb, array[i], len);
			if (rc != 0) return rc;
			array[i][len] = 0;
		}
		*count = nbcount;
	} else if (options & PNETBUFFER_OPT_STRINGS_LAST_EMPTY) {
		i = 0;
		while (1) {
			rc = pnetbuffer_read_uint8(nb, &len);
			if (rc != 0) return rc;
			if (len == 0) break; // empty len, end of array
			if (len >= maxlen) return PNETBUFFER_RC_NOSPACE;
			rc = pnetbuffer_read_data(nb, array[i], len);
			if (rc != 0) return rc;
			array[i][len] = 0;
			i++;
		}
		*count = i;
	} else {
		return PNETBUFFER_RC_MISSINGOPT;
	}

	return rc;
}


static inline uint8_t pnetbuffer_add_string(pnetbuffer_t *nb, char *str)
{
	uint8_t		rc;
	uint16_t	len;

	len = strlen(str);
	rc = pnetbuffer_add_data(nb, str, len);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint8(nb, 0); // NULL terminate
	if (rc != 0) return rc;

	return 0;
}


static inline uint8_t pnetbuffer_read_string(pnetbuffer_t *nb,
					char *str,
					uint16_t strlen)
{
	uint8_t		rc;
	uint16_t	i;

	if (strlen == 0) return PNETBUFFER_RC_NOSPACE;
	i = 0;
	while (1) {
		if (i == strlen) return PNETBUFFER_RC_NOSPACE;
		rc = pnetbuffer_read_uint8(nb, (uint8_t*)&str[i]);
		if (rc != 0) return rc;
		if (str[i] == 0) break;
		i++;
	}

	return 0;
}


static inline uint8_t pnetbuffer_peek_uint8(pnetbuffer_t *nb, uint8_t *val)
{
	if ((nb->rpos) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	*val = nb->buf[nb->rpos];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_peek_uint16(pnetbuffer_t *nb, uint16_t *val)
{
	int		i;

	if ((nb->rpos+1) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	i = nb->rpos;
	*val = (nb->buf[i++])<<8;
	*val += nb->buf[i++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_peek_uint32(pnetbuffer_t *nb, uint32_t *val)
{
	int		i;

	if ((nb->rpos+3) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	i = nb->rpos;
	*val = (nb->buf[i++])<<24;
	*val += (nb->buf[i++])<<16;
	*val += (nb->buf[i++])<<8;
	*val += nb->buf[i++];
	return PNETBUFFER_RC_SUCCESS;
}


static inline uint8_t pnetbuffer_peek_uint64(pnetbuffer_t *nb, uint64_t *val)
{
	int		i;

	if ((nb->rpos+7) >= nb->wpos)
		return PNETBUFFER_RC_NOTENOUGH;
	i = nb->rpos;
	*val = ((uint64_t)(nb->buf[i++]))<<56;
	*val += ((uint64_t)(nb->buf[i++]))<<48;
	*val += ((uint64_t)(nb->buf[i++]))<<40;
	*val += ((uint64_t)(nb->buf[i++]))<<32;
	*val += ((uint64_t)(nb->buf[i++]))<<24;
	*val += ((uint64_t)(nb->buf[i++]))<<16;
	*val += ((uint64_t)(nb->buf[i++]))<<8;
	*val += ((uint64_t)nb->buf[i++]);
	return PNETBUFFER_RC_SUCCESS;
}



#endif


#endif


