/************************************************************************/
/* avltree-addl.c - AVLTree Additional source				*/
/************************************************************************/
/*
 *  (c)2021-2023 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdlib.h>
#include <cutil/avltree.h>
#include <avltree-addl.h>


void iter_stack_push(avltree_iter_t *iter, AVLNode *node)
{
//	printf("%s(): pushing node (data=%d)\n", __func__, *(uint32_t *)node->data);
	if (iter->stack_idx == AVLTREE_MAX_DEPTH)
		return;
	iter->stack[iter->stack_idx] = node;
	iter->stack_idx++;
	return;
}


AVLNode *iter_stack_pop(avltree_iter_t *iter)
{
	if (iter->stack_idx) {
		iter->stack_idx--;
		return iter->stack[iter->stack_idx];
	}
	return NULL;
}


avltree_iter_t *avltree_iter_new(AVLTree *tree, int direction)
{
	avltree_iter_t		*iter = NULL;

	iter = calloc(1, sizeof(avltree_iter_t));
	if (!iter)
		return NULL;
	iter->tree = tree;
	iter->dir = direction;
	iter->stack_idx = 0;

	if (!avltree_root(iter->tree))
		return iter;

	// descend to the outer leaf
	iter->node = avltree_root(tree);
	if (iter->dir == AVLTREE_ITER_ASCEND) {
		while (avlnode_left(iter->node)) {
			iter_stack_push(iter, iter->node);
			iter->node = avlnode_left(iter->node);
		}
	} else {
		while (avlnode_left(iter->node)) {
			iter_stack_push(iter, iter->node);
			iter->node = avlnode_left(iter->node);
		}
	}

	return iter;
}


int avltree_iter_free(avltree_iter_t *iter)
{
	free(iter);
	return 0;
}


void *avltree_iter_next(avltree_iter_t *iter)
{
	int			rc;
	void			*data;
	AVLNode			*n;

	if (iter->flags & AVLTREE_ITER_F_COMPLETE)
		return NULL;
	if (!iter->node) {
		if (!avltree_root(iter->tree))
			return NULL; // empty tree
		// descend to the outer leaf
		iter->node = avltree_root(iter->tree);
		if (iter->dir == AVLTREE_ITER_ASCEND) {
			while (avlnode_left(iter->node)) {
				iter_stack_push(iter, iter->node);
				iter->node = avlnode_left(iter->node);
			}
		} else {
			while (avlnode_left(iter->node)) {
				iter_stack_push(iter, iter->node);
				iter->node = avlnode_left(iter->node);
			}
		}
	}

	data = avlnode_data(iter->node);

	if (iter->dir == AVLTREE_ITER_ASCEND) {
		if (avlnode_right(iter->node)) {
			iter_stack_push(iter, iter->node);
			iter->node = avlnode_right(iter->node);
			// descend left as far as possible
			while (avlnode_left(iter->node)) {
				iter_stack_push(iter, iter->node);
				iter->node = avlnode_left(iter->node);
			}

		} else {
			while (1) {
				n = iter_stack_pop(iter);
				if (!n)
					goto avltree_iter_next_end;
				iter->node = n;
				rc = iter->tree->compare(iter->node->data, data);
				if (rc < 0) {
					if (iter->node == avltree_root(iter->tree))
						goto avltree_iter_next_end;
				} else if (rc > 0) {
					break;
				}
				
			}
		}

	}
	if (avlnode_data(iter->node) == data)
		data = NULL; // end of tree reached
	return data;

avltree_iter_next_end:
	iter->node = NULL;
	iter->flags |= AVLTREE_ITER_F_COMPLETE;
	return data;
}





