/************************************************************************/
/* ioloop.c							     */
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <sys/epoll.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

//#define NDEBUG
#include <assert.h>

#include <dlist.h>
#include <ioloop.h>


#define READ_FD			0
#define WRITE_FD		1

#define FD_EMPTY		-1

#define IOLOOP_FLAG_EXITLOOP	1

#ifndef TIMESPEC_TO_TIMEVAL
#define TIMESPEC_TO_TIMEVAL(tv, ts) {		\
        (tv)->tv_sec = (ts)->tv_sec;		\
        (tv)->tv_usec = (ts)->tv_nsec / 1000;	\
}
#endif


// Internal structure for managing timers
#define IOLOOP_TIMER_F_EXPIRED		 0x00000001
//typedef struct _ioloop_timer_t ioloop_timer_t;
struct _ioloop_timer_t {
	dlist_t			 list;

	ioloop_t		*loop;

	uint32_t		flags;
	struct timeval		expire_tv;
	ioloop_timer_callback	cb;
	void			*arg;
};



// forward declarations
void _ioloop_timer_free(ioloop_t *loop, ioloop_timer_t *timer);



ioloop_t *ioloop_open(int options)
{
	int			rc;
	int			flags;
	int			i;
	struct epoll_event	epevent;
	ioloop_t		*loop = NULL;

	loop = calloc(1, sizeof(ioloop_t));
	if (!loop)
		return NULL;
	loop->poll_timeout = IOLOOP_DEFAULT_TIMEOUT;

	loop->timer_max = IOLOOP_TIMER_MAX_DEFAULT;
	dlist_init(&loop->timer_list);
	dlist_init(&loop->timer_free_list);

	// invalidate all fds in the table
	for (i=0 ; i<EPOLL_TABLE_SIZE ; i++) {
		loop->fd_descriptor[i] = FD_EMPTY;
	}

	loop->epoll_pipe[READ_FD] = -1;
	loop->epoll_pipe[WRITE_FD] = -1;
	rc = pipe(loop->epoll_pipe);
	if (rc == -1)
		goto ioloop_open_exit;
	flags = fcntl(loop->epoll_pipe[READ_FD], F_GETFL, 0);
	rc = fcntl(loop->epoll_pipe[READ_FD],
			F_SETFL,
			flags | O_NONBLOCK);
	if (rc == -1)
		goto ioloop_open_exit;

	loop->epoll_fd = epoll_create(EPOLL_TABLE_SIZE);
	if (loop->epoll_fd == -1)
		goto ioloop_open_exit;

	epevent.data.fd = loop->epoll_pipe[READ_FD];
	epevent.events = EPOLLIN;
	rc = epoll_ctl(loop->epoll_fd,
			EPOLL_CTL_ADD,
			loop->epoll_pipe[READ_FD],
			&epevent);
	if (rc == -1)
		goto ioloop_open_exit;

	return loop;

ioloop_open_exit:
	if (loop->epoll_fd != -1)
		(void)close(loop->epoll_fd);
	if (loop->epoll_pipe[0] != -1)
		(void)close(loop->epoll_pipe[0]);
	if (loop->epoll_pipe[1] != -1)
		(void)close(loop->epoll_pipe[1]);
	if (loop)
		free(loop);
	return NULL;
}


int ioloop_close(ioloop_t *loop)
{
	if (loop->epoll_fd != -1)
		(void)close(loop->epoll_fd);
	if (loop->epoll_pipe[0] != -1)
		(void)close(loop->epoll_pipe[0]);
	if (loop->epoll_pipe[1] != -1)
		(void)close(loop->epoll_pipe[1]);
	if (loop)
		free(loop);
	return 0;
}


int ioloop_set_default_io_callback(ioloop_t *loop,
			ioloop_callback_func func)
{
	loop->default_callback = func;
	return 0;
}


int ioloop_register_fd(ioloop_t *loop,
			int fd,
			uint8_t type,
			uint16_t port,
			ioloop_callback_func callback,
			void *callback_arg)
{
	int			rc;
	int			i;
	struct epoll_event	epevent;

	if (loop->fd_count >= POLL_MAX_SOCKETS)
		return ENOSPC;

	// locate empty table slot
	// TODO eliminate this (just hash on the fd)
	for (i=0 ; i<EPOLL_TABLE_SIZE ; i++) {
		if (loop->fd_descriptor[i] == FD_EMPTY)
			break;
	}
	if (i == EPOLL_TABLE_SIZE)
		return ENOSPC;

	epevent.data.u32 = i;	// store fd array offset
	epevent.events = (EPOLLIN | EPOLLERR | EPOLLHUP);
	rc = epoll_ctl(loop->epoll_fd, EPOLL_CTL_ADD, fd, &epevent);
	if (rc == -1)
		return errno;

	loop->fd_descriptor[i] = fd;
	loop->poll_conn_data[i].type = type;
	loop->poll_conn_data[i].port = port;
	loop->fd_callback[i] = callback;
	loop->fd_callback_arg[i] = callback_arg;
	loop->fd_count++;

	return 0;
}


int ioloop_unregister_fd(ioloop_t *loop, int fd)
{
	int			rc;
	int			i;
	int			oset;

	if (loop->fd_count >= POLL_MAX_SOCKETS)
		return ENOSPC;

	// locate the fd in the array
	// TODO eliminate this (just hash on the fd)
	for (i=0 ; i<loop->fd_count ; i++) {
		if (loop->fd_descriptor[i] == fd)
			break;
	}
	if (i == loop->fd_count)
		return ENOENT;
	oset = i;

	// remove the fd from epoll
	rc = epoll_ctl(loop->epoll_fd, EPOLL_CTL_DEL, fd, NULL);
	if (rc == -1) {
		rc = errno;
		printf("ERROR %s() epoll_ctl() fd=%d rc=%d\n",
				__func__,
				fd,
				rc);
		return rc;
	}

	// clear the entry
	loop->fd_descriptor[oset] = -1;
	loop->poll_conn_data[oset].type = 0;
	loop->poll_conn_data[oset].port = 0;
	loop->fd_callback[oset] = NULL;
	loop->fd_callback_arg[oset] = NULL;
	// update the table entry count
	loop->fd_count--;

	return 0;
}


int ioloop_modify_fd(ioloop_t *loop,
			int fd,
			ioloop_modify_opt_t option,
			ioloop_modify_arg_t *arg)
{
	int			rc;
	int			ti;
	struct epoll_event	epevent;

	// locate table slot for fd
	// TODO eliminate this (just hash on the fd)
	for (ti=0 ; ti<EPOLL_TABLE_SIZE ; ti++) {
		if (loop->fd_descriptor[ti] == fd)
			break;
	}
	if (ti == EPOLL_TABLE_SIZE)
		return ENOENT;

	switch(option) {
		case IOLOOP_MODIFY_FD_O_EVENTS:
			epevent.events = arg->u.events;
			epevent.data.u32 = ti;
			rc = epoll_ctl(loop->epoll_fd,
					EPOLL_CTL_MOD,
					fd,
					&epevent);
			if (rc == -1)
				return errno;
			break;
		default:
			return EINVAL;
	}
	return 0;
}


int timeval_subtract (struct timeval *result,
			struct timeval *x,
			struct timeval *y)
{
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait.
	   tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;

	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}


int ioloop_run(ioloop_t *loop, int options, uint32_t timeout)
{
	int			rc;
	int			i;
	int			efd;
	int			count;
	int			err_count;
	int			sleep_ms;
	char			c;
	ioloop_callback_func	cb;
	struct timespec		curr_ts;
	struct timeval		curr_tv;
	struct timeval		diff_tv;
	ioloop_timer_t		*timer;
	struct epoll_event      epevent[EPOLL_TABLE_SIZE];
	
	loop->flags &= ~(IOLOOP_FLAG_EXITLOOP);
	err_count = 0;
	loop->poll_timeout = timeout;
	while (1) {

		if (loop->flags & IOLOOP_FLAG_EXITLOOP)
			break;

		sleep_ms = loop->poll_timeout;
		if (!dlist_is_empty(&loop->timer_list)) {
			timer = (ioloop_timer_t*)loop->timer_list.next;
			rc = clock_gettime(CLOCK_MONOTONIC, &curr_ts);
			TIMESPEC_TO_TIMEVAL(&curr_tv, &curr_ts);
			rc = timeval_subtract(&diff_tv,
						&timer->expire_tv,
						&curr_tv);
			if (rc == 1) {
				sleep_ms = 1; // expire ASAP
			} else {
				sleep_ms = diff_tv.tv_sec * 1000;
				sleep_ms += diff_tv.tv_usec / 1000;
			}
		}
		if (sleep_ms < 1) {
			printf("%s() WARN changing sleep to 1ms (was 0ms)\n",
				__func__);
			sleep_ms = 1;
			err_count++;
			assert(err_count < 5);
		}

		rc = epoll_wait(loop->epoll_fd,
				epevent,
				EPOLL_TABLE_SIZE,
				sleep_ms);
		if (rc == -1) {
			if (errno == EINTR)
				continue;
			return errno;
		} else if (rc == 0) {
			// handle timeout and continue
			if (!dlist_is_empty(&loop->timer_list)) {
				timer = (ioloop_timer_t*)loop->timer_list.next;
				timer->flags |= IOLOOP_TIMER_F_EXPIRED;
				if (timer->cb) {
					(*timer->cb)(timer, timer->arg);
				}
				if (timer->flags & IOLOOP_TIMER_F_EXPIRED) {
					ioloop_timer_remove(timer);
				}
			}
			// reset err_count
			err_count = 0;
			continue;
		}

		count = rc;
		if (count > 1) {
			printf("%s() DEBUG epoll_wait() count=%d\n",
				__func__,
				count);
		}
		for (i=0 ; i<count ; i++) {

			efd = epevent[i].data.u32;
			if ((epevent[i].events & EPOLLERR)) {
				printf("%s() DEBUG EPOLLERR fd=%d events=%u\n",
					__func__,
					loop->fd_descriptor[efd],
					epevent[i].events);
				err_count++;
				assert(err_count < 5);
			}

			// bad fd, shouldn't happen but....
			if (loop->fd_descriptor[efd] == -1) {
				printf("ERROR %s() bad file descriptor"
					" in table fd=%d\n",
					__func__,
					loop->fd_descriptor[efd]);
				err_count++;
				assert(err_count < 5);
			}

			// handle command pipe
			if (epevent[i].events & EPOLLIN &&
			    loop->fd_descriptor[efd] == loop->epoll_pipe[READ_FD]) {
				rc = read (loop->epoll_pipe[READ_FD], &c, 1);
				if (c == 'E') // loop exit triggered
					return 0;
				continue;
			}

			// handle application fds
			cb = NULL;
			if (loop->fd_callback[efd] != NULL)
				cb = loop->fd_callback[efd];
			else if (loop->default_callback != NULL)
				cb = loop->default_callback;
			if (cb) {
				rc = (*cb)(loop->fd_descriptor[efd],
					epevent[i].events,
					&loop->poll_conn_data[efd],
					loop->fd_callback_arg[efd]);
				(void)rc;
			} else {
				printf("ERROR %s() NULL callback for fd=%d\n",
					__func__,
					loop->fd_descriptor[efd]);
				err_count++;
				if (err_count >= 5)
					return 2;
			}
		}
	}
	return 0;
}


int ioloop_exit(ioloop_t *loop)
{
	int	rc;
	loop->flags |= IOLOOP_FLAG_EXITLOOP;
	rc = write (loop->epoll_pipe[WRITE_FD], "E", 1);
	return rc;
}




//
// timer functions
//


ioloop_timer_t *_ioloop_timer_alloc(ioloop_t *loop)
{
	dlist_t		*item;
	ioloop_timer_t	*timer;

	// DEBUG ONLY
	int i = 0;
	dlist_for_each(item, &loop->timer_free_list) {
		timer = (ioloop_timer_t*)item;
		i++;
		if (i >16) {
			exit(2);
		}
	}

	// allocate from the free list
	if (!dlist_is_empty(&loop->timer_free_list)) {
		timer = (ioloop_timer_t*)loop->timer_free_list.next;
		// leave the existing timer->id intact
		dlist_remove((dlist_t*)timer);
		return timer;
	}

	// ... or allocate a new timer structure
	if (loop->timer_nid >= loop->timer_max) {
		errno = E2BIG;
		return NULL;
	}
	timer = calloc(1, sizeof(ioloop_timer_t));
	if (!timer) {
		errno = ENOMEM;
		return NULL;
	}
	timer->loop = loop;
	loop->timer_nid++;

	return timer;
}


void _ioloop_timer_free(ioloop_t *loop, ioloop_timer_t *timer)
{
	if (timer) {
		timer->flags = 0;
		timer->cb = NULL;
		timer->arg = NULL;
		dlist_add(&loop->timer_free_list,
			(dlist_t*)timer);
	}
	return;
}



int _ioloop_timer_deschedule(ioloop_timer_t *timer)
{
	dlist_t		*item;
	ioloop_timer_t	*t;

	if (dlist_is_empty(&timer->loop->timer_list))
		return 0;

	dlist_for_each(item, &timer->loop->timer_list) {
		t = (ioloop_timer_t*)item;
		if (t == timer) {
			dlist_remove((dlist_t*)timer);
			return 1;
		}
	}
	return 0;
}



void timer_list_insert_by_expire(dlist_t *head, ioloop_timer_t *timer)
{
	dlist_t		*item;
	ioloop_timer_t	*t;

	// add timer to the head
	if (dlist_is_empty(head)) {
		dlist_add(head, (dlist_t*)timer);
		return;
	}

	// ... or add timer in the list
	dlist_for_each(item, head) {
		t = (ioloop_timer_t*)item;
		if (timer->expire_tv.tv_sec < t->expire_tv.tv_sec) {
			dlist_add_tail(item, (dlist_t*)timer);
			return;
		}
		if (timer->expire_tv.tv_sec == t->expire_tv.tv_sec) {
			if (timer->expire_tv.tv_usec < t->expire_tv.tv_usec) {
				dlist_add_tail(item, (dlist_t*)timer);
				return;
			}
		}
	}

	// ... or add timer to the tail
	dlist_add_tail(head, (dlist_t*)timer);

	return;
}


//returns:
//  non-NULL a timer handle to be used in subsequent
//           calls for this timer
//  NULL     on error and errno is set appropriately
ioloop_timer_t *ioloop_timer_add(ioloop_t *loop,
		int expire_ms,
		ioloop_timer_callback cb,
		void *arg)
{
	long		l;
	ioloop_timer_t	*timer;
	struct timespec	ts;

	timer = _ioloop_timer_alloc(loop);
	if (!timer) {
		errno = ENOMEM;
		return NULL;
	}
	clock_gettime(CLOCK_MONOTONIC, &ts);
	TIMESPEC_TO_TIMEVAL(&timer->expire_tv, &ts);
	if (expire_ms >= 1000) {
		l = expire_ms / 1000;
		timer->expire_tv.tv_sec += l;
		expire_ms -= l * 1000;
	}
	timer->expire_tv.tv_usec += (expire_ms / 1000);
	timer->cb = cb;
	timer->arg = arg;
	timer_list_insert_by_expire(&loop->timer_list, timer);
	return timer;
}


int ioloop_timer_remove(ioloop_timer_t *timer)
{
	if (timer) {
		(void)_ioloop_timer_deschedule(timer);
		_ioloop_timer_free(timer->loop, timer);
		return 0;
	}
	errno = ENOENT;
	return -1;
}


int ioloop_timer_enable(ioloop_timer_t *timer)
{
	timer_list_insert_by_expire(&timer->loop->timer_list,
				timer);
	return 0;
}


int ioloop_timer_disable(ioloop_timer_t *timer)
{
	(void)_ioloop_timer_deschedule(timer);
	return 0;
}


int ioloop_timer_set_expiration(ioloop_timer_t *timer,
			int expire_ms)
{
	int		l;
	struct timespec	ts;

	if (!timer) {
		printf("%s() ERROR invalid timer handle %p\n",
			__func__,
			timer);
		exit(2);
		return EINVAL;
	}
	clock_gettime(CLOCK_MONOTONIC, &ts);
	TIMESPEC_TO_TIMEVAL(&timer->expire_tv, &ts);
	if (expire_ms >= 1000) {
		l = expire_ms / 1000;
		timer->expire_tv.tv_sec += l;
		expire_ms -= l * 1000;
	}
	timer->expire_tv.tv_usec += (expire_ms / 1000);
	timer->flags &= ~(IOLOOP_TIMER_F_EXPIRED);
	// remove and reinsert the timer
	dlist_remove((dlist_t*)timer);
	timer_list_insert_by_expire(&timer->loop->timer_list, timer);
	return 0;
}


int ioloop_timer_set_arg(ioloop_timer_t *timer,
			void *arg)
{
	return 0;
}


int ioloop_timer_set_callback(ioloop_timer_t *timer,
			ioloop_timer_callback cb)
{
	return 0;
}



