/*
 * logging.c
 *
 * (c)2017,2020-2021 Todd Shadburn
 *
 * Licensed under the GNU GPL version 2
 *
 *
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <logging.h>




char *logging_level_str[] = {
	"EMERG",
	"ALERT",
	"CRIT",
	"ERROR",
	"WARNING",
	"NOTICE",
	"INFO",
	"DEBUG",
};


char *logging_level_config_str[] = {
	"emergency",
	"alert",
	"critical",
	"error",
	"warning",
	"notice",
	"info",
	"debug",
};



logging_t *logging_open(char *filepath, uint32_t options)
{
	int		rc;
	logging_t	*log = calloc(1, sizeof(logging_t));
	if (!log)
		return NULL;
	log->level = LOG_LEVEL_DEBUG;
	if (options & LOGGING_OPT_MUTEX) {
		rc = pthread_mutex_init(&log->mutex, NULL);
		if (rc != 0)
			goto logging_open_err1;
		log->options |= LOGGING_OPT_MUTEX;
	}
	if (options & LOGGING_OPT_FLUSH_LOG_ENTRIES)
		log->options |= LOGGING_OPT_FLUSH_LOG_ENTRIES;
	if (filepath) {
		strcpy(log->filepath, filepath);
		log->fp = fopen(log->filepath, "a");
		if (!log->fp)
			goto logging_open_err1;
	} else {
		log->fp = stdout;
	}

	return log;

logging_open_err1:
	free(log);
	return NULL;
}


int logging_close(logging_t *log)
{
	int rc;
	if (log->fp != stdout) {
		rc = fclose(log->fp);
		(void)rc; // avoid compiler wanrings
	}
	free(log);
	return 0;
}


int logging_reopen(logging_t *log, char *filepath, uint32_t options)
{
	int		rc;

	if (log->fp != stdout)
		rc = fclose(log->fp);
	if (filepath) {
		strcpy(log->filepath, filepath);
		log->fp = fopen(log->filepath, "a");
		if (!log->fp) {
			rc = errno;
			log->fp = stdout;
			goto logging_reopen_err1;
		}
	} else {
		log->fp = stdout;
	}
	if (options & LOGGING_OPT_FLUSH_LOG_ENTRIES)
		log->options |= LOGGING_OPT_FLUSH_LOG_ENTRIES;
	return 0;

logging_reopen_err1:
	return rc;
}


void logging_entry_va(logging_t *log, int level, char *format, va_list args)
{
//	va_list		args;
	int		rc;
	int		ll;
	time_t		t;
	struct tm	*lt = NULL;
	if (log == NULL || format == NULL)
		return;
	ll = level;
	if (level < LOG_LEVEL_EMERG || level >= _LOG_LEVEL_MAX)
		ll = LOG_LEVEL_DEBUG;
	if (ll > log->level)
		return; // dont log above current level
//	va_start (args, format);
	t = time (NULL);
	lt = localtime (&t);
	if (log->options & LOGGING_OPT_MUTEX)
		pthread_mutex_lock (&log->mutex);
	rc = vsnprintf (log->buffer,
			LOGGING_BUFFER_SIZE,
			format,
			args);
	if (rc > 0) {
		fprintf (log->fp,
			"%04d-%02d-%02dT%02d:%02d:%02d %s %s\n",
			lt->tm_year+1900,
			lt->tm_mon+1,
			lt->tm_mday,
			lt->tm_hour,
			lt->tm_min,
			lt->tm_sec,
			logging_level_str[ll],
			log->buffer);
	}
	if (log->options & LOGGING_OPT_MUTEX)
		pthread_mutex_unlock (&log->mutex);
	if (log->options & LOGGING_OPT_FLUSH_LOG_ENTRIES)
		fflush (log->fp);
//	va_end (args);
	return;
}


void logging_entry(logging_t *log, int level, char *format, ...)
{
	va_list		args;

	va_start(args, format);
	logging_entry_va(log, level, format, args);
	va_end(args);
	return;
}


void logging_set_log_level(logging_t *log, int level)
{
	int	ll;
	ll = level;
	if (level < LOG_LEVEL_EMERG || level >= _LOG_LEVEL_MAX)
		ll = LOG_LEVEL_DEBUG;
	log->level = ll;
	return;
}


logging_level_e logging_get_log_level(logging_t *log)
{
	return log->level;
}


int logging_string_to_level(char* str)
{
	int		i;
	for (i=0 ; i< _LOG_LEVEL_MAX; i++) {
		if (!strcmp(str, logging_level_config_str[i]))
			return i;
	}
	return LOG_LEVEL_DEBUG;
}


char *logging_level_to_string(int level)
{
	if (level < LOG_LEVEL_EMERG || level >= _LOG_LEVEL_MAX)
		return NULL;
	return logging_level_config_str[level];
}



