/************************************************************************/
/* v4cidr.c - v4cidr defines						*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/slist.h>

#include <v4cidr.h>



#define CIDR_STRING_SIZE	32

#define UINT32_UPPER(x) ((x & 0xffff0000)>>16)


v4cidr_hashtable_t *v4cidr_hashtable_new(int width)
{
	v4cidr_hashtable_t	*ht = NULL;
	ht = calloc(1, sizeof(v4cidr_hashtable_t));
	if (!ht)
		return NULL;
	if (width < 256 || width > V4CIDR_TABLE_WIDTH_DEFAULT)
		ht->width = V4CIDR_TABLE_WIDTH_DEFAULT;
	else
		ht->width = width;
	ht->table = calloc(ht->width, sizeof(SList));
	if (!ht->table) {
		free(ht);
		return NULL;
	}
	return ht;
}


int v4cidr_hashtable_free(v4cidr_hashtable_t *table)
{
	if (!table)
		return EINVAL;
	if (table->table) {
		// TODO free entry lists
		free(table->table);
	}
	free(table);
	return 0;
}


int cidr2entry(char *cidrpair, v4cidr_entry_t *entry)
{
	int		rc;
	char		*p;
	char		*t;
	struct in_addr	addr;
	char		cp[CIDR_STRING_SIZE];

	// a copy for strtok_p to perturb
	strncpy(cp, cidrpair, CIDR_STRING_SIZE);
	cp[CIDR_STRING_SIZE-1] = 0;

	p = strtok_r(cp, "/", &t);
	if (!p) return EINVAL;
	rc = inet_aton(p, &addr);
	if (!rc) return EINVAL;
	entry->net = ntohl(addr.s_addr);

	p = strtok_r(NULL, "\r\n", &t);
	if (!p) return EINVAL;
	rc = inet_aton(p, &addr);
	if (!rc) return EINVAL;
	entry->mask = ntohl(addr.s_addr);

	return 0;
}


int v4cidr_hashtable_add_cidr(v4cidr_hashtable_t *table,
				char *cidr,
				void *data)
{
	int		rc;
	int		h;
	v4cidr_entry_t	*entry;
	v4cidr_entry_t	*e;
	SList		*l;

	entry = calloc(1, sizeof(v4cidr_entry_t));
	if (!entry) return ENOMEM;
	rc = cidr2entry(cidr, entry);
	if (rc) return rc;
	entry->data = data;
	h = UINT32_UPPER(entry->net);
	l = table->table[h];
	while (l) {
		e = slist_get_data(l);
		if (e->net == entry->net) return EEXIST;
		l = slist_get_next(l);
	}
	rc = slist_item_add(&table->table[h], entry);

	return rc;
}


int v4cidr_hashtable_remove_cidr(v4cidr_hashtable_t *table,
				char *cidr)
{
	int		rc;
	int		h;
	v4cidr_entry_t	entry;
	v4cidr_entry_t	*e;
	SList		*l;

	rc = cidr2entry(cidr, &entry);
	if (rc) return rc;
	h = UINT32_UPPER(entry.net);
	l = table->table[h];
	while (l) {
		e = slist_get_data(l);
		if (e->net == entry.net) break;
		l = slist_get_next(l);
	}
	if (!l) return ENOENT;
	rc = slist_item_remove(&table->table[h], e);
	free(e);

	return rc;
}


const v4cidr_entry_t *v4cidr_hashtable_locate(v4cidr_hashtable_t *table,
						char *ip)
{
	int		rc;
	int		h;
	struct in_addr	addr;
	v4cidr_entry_t	*e;
	SList		*l;

	rc = inet_aton(ip, &addr);
	if (!rc) return NULL;
	addr.s_addr = ntohl(addr.s_addr);

	h = UINT32_UPPER(addr.s_addr);
	l = table->table[h];
	while (l) {
		e = slist_get_data(l);
		if ( (addr.s_addr&e->mask) == e->net)
			return e;
		l = slist_get_next(l);
	}

	return NULL;
}


void *v4cidr_hashtable_locate_get_data(v4cidr_hashtable_t *table,
                                                char *ip)
{
	const v4cidr_entry_t	*e;

	e = v4cidr_hashtable_locate(table, ip);
	if (e)
		return e->data;

	return NULL;
}


int v4cidr_hashtable_exists(v4cidr_hashtable_t *table,
			char *ip)
{
	const v4cidr_entry_t	*e;

	e = v4cidr_hashtable_locate(table, ip);
	if (e)
		return 1;

	return 0;
}



